from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generator,
    Generic,
    List,
    Optional,
    Sequence,
    Set,
    Type,
    TypeVar,
    Union,
    cast,
)

from chalk.utils.collections import ensure_tuple, get_unique_item
from chalk.utils.duration import Duration

if TYPE_CHECKING:
    from chalk.features.feature import Feature

T = TypeVar("T")


class FeatureType(Generic[T]):
    def __init__(
        self,
        name: Optional[str],
        namespace: str,
        encoder: Optional[Callable[[T], Any]],
        decoder: Optional[Callable[[Any], T]],
        typ: Optional[Union[DataFrameType, Type[T]]],
        is_nullable: bool,
    ) -> None:
        self.name = name
        self.namespace = namespace
        self.encoder = encoder
        self.decoder = decoder
        self.type = typ
        self.is_nullable = is_nullable

    def is_scalar_primary(self) -> bool:
        return False

    def root_namespace(self) -> str:
        """The first component of the namespace.

        If the feature is an input feature, the root namespace should be the root namespace of the resolver.
        For example, if a resolver referenced ``Child.parent.name``, then this would be ``child``.

        Otherwise, this is equivalent to :attr:`namespace`.
        """
        return self.namespace

    def root_prefix(self) -> str:
        """The full namespace path, relative to the root, without the feature name.

        If the feature is an input feature, the rooted prefix will begin with the root namespace of the resolver,
        and contain the feature names of all relationships (except for the feature name in foreign features class).
        For example, if a resolver referenced ``Child.parent.name``, then this would be ``child.parent``.

        Otherwise, this is equivalent to :attr:`namespace`.
        """
        return self.namespace

    def root_fqn(self) -> str:
        """The fqn, relative to the root namespace.

        If the feature is an input feature, the rooted fqn will begin with the root namespace of the resolver,
        and contain all feature names of all relationships, including the feature name.
        For example, if a resolver referenced ``Child.parent.name``, then this would be ``child.parent.name``.

        Otherwise, this is equivalent to :attr:`fqn`.
        """
        return self.fqn

    @staticmethod
    def of(feature: Union[str, Feature, FeatureType]) -> "FeatureType":
        if isinstance(feature, FeatureType):
            return feature

        from chalk.parsed.conversions import DEFAULT_CONVERTER

        if isinstance(feature, str):
            raw_str = feature
            converted_feature = DEFAULT_CONVERTER.feature_from_string(raw_str)
            if converted_feature is None:
                raise ValueError(f'Could not resolve feature "{raw_str}"')
            feature = converted_feature

        return DEFAULT_CONVERTER.convert_target(feature)

    @property
    def fqn(self):
        if self.name is None:
            return self.namespace
        else:
            return f"{self.namespace}.{self.name}"

    def __str__(self):
        return self.root_fqn()

    def __repr__(self):
        return f"{type(self).__name__}(fqn={self.fqn}, type={str(self.type)})"

    def __hash__(self) -> int:
        return hash(self.fqn)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FeatureType):
            return NotImplemented
        if not isinstance(other, type(self)):
            return False
        return self.fqn == other.fqn


class InputFeatureType(FeatureType[T]):
    """FeatureType used to represent the children of has-one relationships. For"""

    def __init__(self, underlying: FeatureType[T], path: List[HasOnePathObjParsed]):
        super().__init__(
            name=underlying.name,
            namespace=underlying.namespace,
            encoder=None,
            decoder=None,
            typ=None,
            is_nullable=False,
        )
        assert not isinstance(
            underlying, InputFeatureType
        ), "Recursive InputFeaturetypes are disallowed. Instead, update the path."
        self.underlying = underlying
        self.path = path

    def walk_path_joins(self) -> Generator["HasOneFeatureType", Any, Any]:
        for x in self.path:
            yield x.parent

    def root_namespace(self) -> str:
        return self.path[0].parent.namespace

    def root_prefix(self) -> str:
        return self.root_namespace() + "." + ".".join(cast(str, x.parent.name) for x in self.path)

    def root_fqn(self):
        assert self.name is not None
        return self.root_prefix() + "." + self.name

    def __repr__(self):
        return f"InputFeatureType(fqn={self.underlying.fqn}, root_prefix={self.root_prefix()})"

    def __eq__(self, other):
        return isinstance(other, InputFeatureType) and other.underlying == self.underlying and self.path == other.path

    def __hash__(self):
        return self.underlying.__hash__()

    @property
    def join_path_keys(self) -> Set["ScalarFeatureType"]:
        ret = set()
        for p in self.path:
            ret.update(p.parent.join.referenced_features)
        return ret

    @property
    def join_path_spine(self) -> Set["ScalarFeatureType"]:
        ret = set()
        for p in self.path:
            ret.update([x for x in p.parent.join.referenced_features if x.namespace == p.parent.namespace])
        return ret

    @property
    def max_staleness(self):
        if isinstance(self.underlying, (ScalarFeatureType, InputFeatureType)):
            return self.underlying.max_staleness
        raise RuntimeError("Max staleness is not applicable")


class FeatureTimeFeatureType(FeatureType[datetime]):
    def __init__(self, name: str, namespace: str):
        super().__init__(
            name=name,
            namespace=namespace,
            encoder=None,
            decoder=None,
            typ=datetime,
            is_nullable=False,
        )


class ScalarFeatureType(FeatureType[T]):
    def is_scalar_primary(self) -> bool:
        return self.primary

    def __init__(
        self,
        name: str,
        namespace: str,
        type: Type[T],
        is_nullable: bool,
        is_list: bool,
        primary: bool = False,
        description: Optional[str] = None,
        version: Optional[int] = None,
        owner: Optional[str] = None,
        tags: Optional[Union[str, List[str]]] = None,
        max_staleness: Optional[Duration] = None,
        etl_offline_to_online: bool = False,
        encoder: Optional[Callable[[T], Any]] = None,
        decoder: Optional[Callable[[Any], T]] = None,
    ):
        super().__init__(
            namespace=namespace,
            name=name,
            encoder=encoder,
            decoder=decoder,
            typ=type,
            is_nullable=is_nullable,
        )
        self.is_list = is_list
        self.description = description
        self.owner = owner
        self.version = version
        self.tags = ensure_tuple(tags)
        self.primary = primary
        self.max_staleness = max_staleness
        self.etl_offline_to_online = etl_offline_to_online

    def required_by_join_path(self) -> Set[FeatureType]:
        return set()


@dataclass
class FilterParsed:
    lhs: Any
    operation: str
    rhs: Any

    def is_simple_join(self):
        return (
            isinstance(self.lhs, ScalarFeatureType)
            and isinstance(self.rhs, ScalarFeatureType)
            and self.operation == "=="
        )

    # If this filter is resolved relative to timestamps on the root query (or 'now') then we need
    # to handle it per-row in a fancy way
    def is_constant_filter(self) -> bool:
        from chalk.features.feature import TimeDelta

        referenced_time_delta = isinstance(self.lhs, TimeDelta) or isinstance(self.rhs, TimeDelta)
        return not referenced_time_delta

    @property
    def referenced_features(self) -> Set[ScalarFeatureType]:
        return set(v for v in (self.lhs, self.rhs) if isinstance(v, ScalarFeatureType))


class HasOneFeatureType(FeatureType[T]):
    def __init__(self, namespace: str, name: str, join: FilterParsed, is_nullable: bool):
        super().__init__(
            namespace=namespace,
            name=name,
            encoder=None,
            decoder=None,
            typ=None,
            is_nullable=is_nullable,
        )
        self.join = join

    # this will be replaced once support for multiple join conditions is fully implemented
    def only_local_feature(self) -> ScalarFeatureType:
        return self.local_features()[0]

    # this will be replaced once support for multiple join conditions is fully implemented
    def only_foreign_features(self) -> ScalarFeatureType:
        return self.foreign_features()[0]

    def local_features(self) -> List[ScalarFeatureType]:
        return list(filter(lambda x: x.namespace == self.namespace, self.join.referenced_features))

    def foreign_features(self) -> List[ScalarFeatureType]:
        return list(filter(lambda x: x.namespace != self.namespace, self.join.referenced_features))

    def foreign_namespace(self) -> str:
        return next(iter(self.foreign_features())).namespace

    def get_first_level_scalars(self):
        assert isinstance(self.type, FeatureSetType)
        return [f for f in self.type.features if isinstance(f, ScalarishFeatureType)]

    def __eq__(self, other):
        return (
            isinstance(other, HasOneFeatureType)
            and self.name == other.name
            and self.namespace == other.namespace
            and self.join == other.join
        )

    def __hash__(self):
        # FIXME(Andy) (need to come up with a unique hash)
        return hash((self.name, self.namespace))


class DataFrameType(FeatureType[None]):
    def __init__(self, filters: List[FilterParsed], columns: List[QueryInputType]):
        namespaces = [x.root_namespace() for x in columns]
        namespace = get_unique_item(namespaces, name="dataframe column namespace")
        super().__init__(
            encoder=None,
            decoder=None,
            typ=None,
            name=None,
            namespace=namespace,
            is_nullable=False,
        )
        self.filters = filters
        self.columns = columns

    def __repr__(self):
        return f"DataFrameType(namespace={self.namespace}, filters={str(self.filters)}, columns={str(self.columns)})"

    def __hash__(self):
        return hash((tuple(self.filters), tuple(self.columns)))

    def __eq__(self, other):
        if not isinstance(other, DataFrameType):
            return NotImplemented

        return self.filters == other.filters and self.columns == other.columns

    def all_referenced_features(self) -> Set[FeatureType]:
        ret: Set[FeatureType] = set()

        ret.update(self.columns)

        for filter in self.filters:
            if isinstance(filter.lhs, ScalarFeatureType):
                ret.add(filter.lhs)

            if isinstance(filter.rhs, ScalarFeatureType):
                ret.add(filter.rhs)

        return ret


class FeatureSetType(FeatureType[T]):
    def __init__(self, features: Sequence[FeatureType[T]]) -> None:
        namespaces = [f.namespace for f in features]
        namespace = get_unique_item(namespaces, name="feature namespace")
        super().__init__(
            name=None,
            namespace=namespace,
            encoder=None,
            decoder=None,
            typ=None,
            is_nullable=False,
        )
        assert len(features) > 0
        self.features = features

    def __eq__(self, other):
        return isinstance(other, FeatureSetType) and self.features == other.features

    def __hash__(self):
        return hash(tuple(self.features))

    def is_single_data_frame(self) -> bool:
        return len(self.features) == 1 and isinstance(self.features[0], DataFrameType)

    def get_first_level_scalars(self):
        return [f for f in self.features if isinstance(f, ScalarishFeatureType)]


class HasManyFeatureType(FeatureType[DataFrameType]):
    join: FilterParsed

    def __init__(
        self,
        namespace: str,
        name: str,
        join: FilterParsed,
        type: Optional[DataFrameType],
    ):
        super().__init__(
            namespace=namespace,
            name=name,
            encoder=None,
            decoder=None,
            typ=type,
            is_nullable=False,
        )
        self.join = join

    def __eq__(self, other):
        if not isinstance(other, HasManyFeatureType):
            return NotImplemented

        namespace_eq = self.namespace == other.namespace
        name_eq = self.name == other.name

        if namespace_eq and name_eq:
            # trivial perf opt...
            type_eq = self.type == other.type
            join_eq = self.join == other.join
            return type_eq and join_eq
        else:
            return False

    def __hash__(self) -> int:
        return super().__hash__()

    def local_features(self) -> List[FeatureType]:
        return list(filter(lambda x: x.namespace == self.namespace, self.join.referenced_features))

    def foreign_features(self) -> List[FeatureType]:
        return list(filter(lambda x: x.namespace != self.namespace, self.join.referenced_features))

    def foreign_namespace(self) -> str:
        return next(
            filter(lambda x: x.namespace != self.namespace, self.join.referenced_features),
        ).namespace


JoinFeatureType = Union[HasOneFeatureType, HasManyFeatureType]


@dataclass
class HasOnePathObjParsed:
    parent: HasOneFeatureType
    # The final child in a path will not be a HasOneFeatureType,
    # and all other intermediate children will be HasOneFeatureType
    child: FeatureType
    parent_to_child_attribute_name: str

    def __hash__(self):
        return hash((self.parent.fqn, self.parent_to_child_attribute_name))

    def __eq__(self, other):
        return (
            isinstance(other, HasOnePathObjParsed)
            and self.parent_to_child_attribute_name == other.parent_to_child_attribute_name
            and self.parent == other.parent
            and self.child == other.child
        )


ScalarishFeatureType = Union[ScalarFeatureType, FeatureTimeFeatureType]
QueryInputType = Union[ScalarishFeatureType, HasManyFeatureType, HasOneFeatureType, InputFeatureType]
QueryOutputType = Union[ScalarishFeatureType, HasManyFeatureType, InputFeatureType]
TIMESTAMP = ScalarFeatureType(
    namespace="global__",
    name="global_timestamp__",
    description="Timestamp Field",
    primary=False,
    type=datetime,
    is_nullable=False,
    is_list=False,
)



import logging
import time
from collections import defaultdict
from enum import Enum
from typing import Any, List, Mapping, Optional, Union

import numpy as np
import pandas as pd
from sqlalchemy import text

from chalk.features import Feature, Features
from chalk.parsed.feature import FeatureType
from chalk.sql.base.protocols import (
    BaseSQLSourceProtocol,
    ChalkQueryProtocol,
    DBSessionProtocol,
    IncrementalSettings,
    StringChalkQueryProtocol,
)


class Finalizer(str, Enum):
    OneOrNone = "OneOrNone"
    One = "One"
    First = "First"
    All = "All"


_logger = logging.getLogger(__name__)


class StringChalkQuery(StringChalkQueryProtocol):
    def __init__(
        self,
        session: DBSessionProtocol,
        source: BaseSQLSourceProtocol,
        query: str,
        fields: Mapping[str, Union[Feature, str]],
        args: Optional[Mapping[str, str]],
    ):
        self._finalizer: Optional[Finalizer] = None
        self._session = session
        self._source = source
        self._query = text(query)
        self._fields = fields
        self._incremental_settings: Optional[Union[IncrementalSettings, bool]] = None
        if args is not None:
            self._query = self._query.bindparams(**args)

    def one_or_none(self):
        self._finalizer = Finalizer.OneOrNone
        return self

    def one(self):
        self._finalizer = Finalizer.One
        return self

    def all(self, incremental: Union[bool, IncrementalSettings] = False):
        self._finalizer = Finalizer.All
        if isinstance(incremental, bool) and incremental:
            self._incremental_settings = IncrementalSettings(lookback_period=None)
        elif isinstance(incremental, IncrementalSettings):
            self._incremental_settings = incremental
        return self

    def execute(self):
        tuples = self._session.execute(self._query)
        cols = [FeatureType.of(v) for (k, v) in self._fields.items()]
        self._finalizer = self._finalizer or Finalizer.All

        if self._finalizer == Finalizer.All:
            data = [list(tup[field] for (field, _) in self._fields.items()) for tup in tuples]
            df = pd.DataFrame(
                data=data,
                columns=cols,
            )
            df.replace({np.nan: None}, inplace=True)
            from chalk.df.ChalkDataFrameImpl import ChalkDataFrameImpl

            return ChalkDataFrameImpl(df)

        fs = Features[cols]()
        tups = list(tuples)
        tup = None
        if self._finalizer == Finalizer.One:
            if len(tups) == 1:
                tup = tups[0]
            else:
                raise ValueError("Expected exactly one result")

        if self._finalizer == Finalizer.First:
            if len(tups) > 1:
                tup = tups[0]
            else:
                raise ValueError("Expected at least one result")

        if self._finalizer == Finalizer.OneOrNone and len(tups) > 0:
            tup = tups[0]

        if tup:
            for (field, feature) in self._fields.items():
                assert isinstance(feature, (FeatureType, Feature))
                assert feature.name is not None
                setattr(fs, feature.name, tup[field])
        return fs


class ChalkQuery(ChalkQueryProtocol):
    _session: DBSessionProtocol
    _source: BaseSQLSourceProtocol
    _features: List[Feature]
    _finalizer: Optional[Finalizer]
    _incremental_settings: Optional[Union[IncrementalSettings, bool]]

    def __init__(
        self,
        features: List[Feature],
        session: DBSessionProtocol,
        source: BaseSQLSourceProtocol,
        raw_session: Optional[Any] = None,
    ):
        self._session = session
        self._raw_session = raw_session or session
        self._features = features
        self._finalizer = None
        self._source = source
        self._incremental_settings = None

    def first(self):
        self._session.update_query(lambda x: x.limit(1))
        self._finalizer = Finalizer.First
        return self

    def one_or_none(self):
        self._session.update_query(lambda x: x.limit(1))
        self._finalizer = Finalizer.OneOrNone
        return self

    def one(self):
        self._session.update_query(lambda x: x.limit(1))
        self._finalizer = Finalizer.One
        return self

    def all(self, incremental: Union[bool, IncrementalSettings] = False):
        self._finalizer = Finalizer.All
        if isinstance(incremental, bool) and incremental:
            self._incremental_settings = IncrementalSettings(lookback_period=None)
        elif isinstance(incremental, IncrementalSettings):
            self._incremental_settings = incremental
        return self

    def filter_by(self, **kwargs):
        self._session.update_query(lambda x: x.filter_by(**kwargs))
        return self

    def filter(self, *criterion):
        self._session.update_query(lambda x: x.filter(*criterion))
        return self

    def limit(self, *limits):
        self._session.update_query(lambda x: x.limit(*limits))
        return self

    def order_by(self, *clauses):
        self._session.update_query(lambda x: x.order_by(*clauses))
        return self

    def group_by(self, *clauses):
        self._session.update_query(lambda x: x.group_by(*clauses))
        return self

    def having(self, criterion):
        self._session.update_query(lambda x: x.having(*criterion))
        return self

    def union(self, *q):
        self._session.update_query(lambda x: x.union(*q))
        return self

    def union_all(self, *q):
        self._session.update_query(lambda x: x.union_all(*q))
        return self

    def intersect(self, *q):
        self._session.update_query(lambda x: x.intersect(*q))
        return self

    def intersect_all(self, *q):
        self._session.update_query(lambda x: x.intersect_all(*q))
        return self

    def join(self, target, *props, **kwargs):
        self._session.update_query(lambda x: x.join(target, *props, **kwargs))
        return self

    def outerjoin(self, target, *props, **kwargs):
        self._session.update_query(lambda x: x.outerjoin(target, *props, **kwargs))
        return self

    def select_from(self, *from_obj):
        self._session.update_query(lambda x: x.select_from(*from_obj))
        return self

    @staticmethod
    def _get_finalizer_fn(f: Optional[Finalizer]):
        if f == Finalizer.First:
            return lambda x: x.first()
        if f == Finalizer.All:
            return lambda x: x.all()
        if f == Finalizer.One:
            return lambda x: x.one()
        if f == Finalizer.OneOrNone:
            return lambda x: x.one_or_none()
        if f is None:
            return lambda x: x.all()
        raise ValueError(f"Unknown finalizer {f}")

    def execute(self):
        from chalk.serialization.codec import FeatureCodec

        codec = FeatureCodec()
        try:
            start = time.perf_counter()

            column_descriptions = self._session._session.column_descriptions
            self._session.update_query(self._get_finalizer_fn(self._finalizer))
            tuples = self._session.result()
            self._raw_session.close()
            fqn_to_feature = {f.fqn: f for f in self._features if isinstance(f, (Feature, FeatureType))}
            cols = [
                FeatureType.of(fqn_to_feature[value["name"]])
                for value in column_descriptions
                if value["name"] in fqn_to_feature
            ]

            if isinstance(tuples, list):
                lists = defaultdict(list)

                for tuple in tuples:
                    for col_i, x in enumerate(tuple):
                        if isinstance(x, Enum):
                            x = x.value
                        lists[col_i].append(x)

                df = pd.DataFrame(
                    {col: pd.Series(lists[i], dtype=codec.get_dtype(fqn=col.fqn)) for i, col in enumerate(cols)}
                )

                df.replace({np.nan: None, pd.NaT: None}, inplace=True)
                from chalk.df.ChalkDataFrameImpl import ChalkDataFrameImpl

                return ChalkDataFrameImpl(df)

            fs = Features[cols]()
            if tuples is not None:
                for (feat, value) in zip(fs.features, tuples):
                    assert isinstance(feat, (FeatureType, Feature))
                    assert feat.name is not None
                    setattr(fs, feat.name, value)
            return fs
        finally:
            _logger.debug(f"query.execute: {time.perf_counter() - start}")



from chalk.sql.base.sql_source import BaseSQLSource


class InMemorySQLite(BaseSQLSource):
    def local_engine_url(self) -> str:
        return "sqlite:///:memory:?check_same_thread=true"


class FileSQLite(BaseSQLSource):
    def __init__(self, filename: str):
        self.filename = filename
        super(FileSQLite, self).__init__()

    def local_engine_url(self) -> str:
        return f"sqlite:///{self.filename}?check_same_thread=true"



import typing
from typing import TypeVar

from pydantic import BaseModel
from typing_extensions import Protocol

T = TypeVar("T")


class StreamSource(Protocol[T]):
    Message: typing.Any
    consumer_config: BaseModel



import unittest

from pydantic import BaseModel

from chalk.features import features
from chalk.streams import KafkaSource, stream
from chalk.streams.StreamUpdate import StreamUpdate


@features
class StreamFeatures:
    scalar_feature: str


class KafkaMessage(BaseModel):
    val_a: str


s = KafkaSource(message=KafkaMessage)


@stream
def fn(message: s.Message):
    return StreamUpdate(
        online=StreamFeatures(
            scalar_feature=message.val_a,
        )
    )


class StreamTestCase(unittest.TestCase):
    def test_callable(self):
        self.assertEqual(
            fn(KafkaMessage(val_a="hello")),
            StreamUpdate(online=StreamFeatures(scalar_feature="hello")),
        )

    def test_parsed_source(self):
        self.assertEqual(fn.source, s)



from typing import List, Mapping

import pandas as pd
from sqlalchemy.orm import Query as SqlAlchemyQuery

from chalk.features import Feature
from chalk.features.feature import Features
from chalk.parsed.feature import FeatureType


class ChalkSqlAlchemyQuery:
    _query: SqlAlchemyQuery
    _features: List[Feature]

    def __init__(self, features: List[Feature], query: SqlAlchemyQuery):
        self._query = query
        self._features = features

    def _updated(self, q: SqlAlchemyQuery):
        self._query = q
        return self

    def _all_as_result(self, tup):
        fqn_to_feature: Mapping[str, Feature] = {
            f.fqn: f for f in self._features if isinstance(f, (Feature, FeatureType))
        }
        fs = Features()
        fs.features = tuple(
            (fqn_to_feature[col_description["name"]] for col_description in self._query.column_descriptions)
        )
        if tup is not None:
            for (feat, value) in zip(fs.features, tup):
                setattr(fs, feat.name, value)
        else:
            raise ValueError("Failed to fetch any results from db")
        return fs

    def first(self):
        return self._all_as_result(self._query.first())

    def one_or_none(self):
        return self._all_as_result(self._query.one_or_none())

    def one(self):
        return self._all_as_result(self._query.one())

    def all(self):
        fqn_to_feature = {f.fqn: f for f in self._features if isinstance(f, (Feature, FeatureType))}
        tuples = self._query.all()
        return pd.DataFrame(
            tuples,
            columns=[fqn_to_feature[value["name"]] for value in self._query.column_descriptions],
        )

    def filter_by(self, **kwargs):
        return self._updated(self._query.filter_by(**kwargs))

    def filter(self, *criterion):
        return self._updated(self._query.filter(*criterion))

    def order_by(self, *clauses):
        return self._updated(self._query.order_by(*clauses))

    def group_by(self, *clauses):
        return self._updated(self._query.group_by(*clauses))

    def having(self, criterion):
        return self._updated(self._query.having(*criterion))

    def union(self, *q):
        return self._updated(self._query.union(*q))

    def union_all(self, *q):
        return self._updated(self._query.union_all(*q))

    def intersect(self, *q):
        return self._updated(self._query.intersect(*q))

    def intersect_all(self, *q):
        return self._updated(self._query.intersect_all(*q))

    def join(self, target, *props, **kwargs):
        return self._updated(self._query.join(target, *props, **kwargs))

    def outerjoin(self, target, *props, **kwargs):
        return self._updated(self._query.outerjoin(target, *props, **kwargs))

    def select_from(self, *from_obj):
        return self._updated(self._query.select_from(*from_obj))



from typing import List, Optional, TypeVar, Union

from pydantic import BaseModel

from chalk.streams.StreamSource import StreamSource


class KafkaConsumerConfig(BaseModel):
    broker: Union[str, List[str]]
    topic: Union[str, List[str]]
    ssl_keystore_location: Optional[str] = None
    client_id_prefix: Optional[str] = None
    group_id_prefix: Optional[str] = None
    topic_metadata_refresh_interval_ms: Optional[int] = None
    security_protocol: Optional[str] = None


T = TypeVar("T")


class KafkaSource(StreamSource[T]):
    def __init__(
        self,
        consumer_config: Optional[KafkaConsumerConfig] = None,
        message: Optional[T] = None,
    ):
        super(KafkaSource, self).__init__()
        self.Message = self
        self._message = message
        self.consumer_config = consumer_config



import logging
import os
from typing import Any, Mapping, Optional

from sqlalchemy.engine.url import URL

from chalk.sql.base.sql_source import BaseSQLSource

_logger = logging.getLogger(__name__)


class CloudSQLSource(BaseSQLSource):
    def __init__(
        self,
        instance_name: Optional[str] = None,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        name: Optional[str] = None,
    ):
        prefix = name + "_" if name else ""
        self.instance_name = instance_name or os.getenv(prefix + "CLOUDSQL_INSTANCE_NAME")
        self.db = db or os.getenv(prefix + "CLOUDSQL_DATABASE")
        self.user = user or os.getenv(prefix + "CLOUDSQL_USER")
        self.password = password or os.getenv(prefix + "CLOUDSQL_PASSWORD")

        super(CloudSQLSource, self).__init__()

    def engine_args(self) -> Mapping[str, Any]:
        return dict(
            pool_size=20,
            max_overflow=60,
            # Trying to fix mysterious dead connection issue
            connect_args={
                "keepalives": 1,
                "keepalives_idle": 30,
                "keepalives_interval": 10,
                "keepalives_count": 5,
            },
        )

    def local_engine_url(self) -> str:
        _logger.debug(f"Connecting to CloudSQL instance {self.instance_name}")
        return URL.create(
            drivername="postgresql",
            username=self.user,
            password=self.password,
            host="",
            query={"host": "{}/{}/.s.PGSQL.5432".format("/cloudsql", self.instance_name)},
            database=self.db,
        )



import unittest
from datetime import datetime

from chalk.features import feature, feature_time, features
from chalk.features.feature import unwrap_feature


@features(tags="t", owner="andy@chalk.ai")
class WowFS:
    # This is a really neat description of something
    something: str

    # This is also really neat and cool
    something_else: str
    nocomment: str

    nope: str

    assigned: str = feature(tags="a")

    # bizarre
    bizarre: str

    # goofy
    goofy: str = feature(owner="yo")

    # now with feature
    assigned_comment: str = feature(tags=["3", "4"])

    # implicit comment
    explicit: str = feature(description="explicit comment")

    # Multiline
    # Neat and, verily, cool
    #
    # Hello
    assigned_comment_multiline: str = feature()

    # now with feature time
    time: datetime = feature_time()

    nope_nope: datetime  # Datetime field that is not feature time


@features(owner="elliot@chalk.ai")
class OwnerFeatures:
    plain: str
    cached: str = feature(max_staleness="3d")
    andy: str = feature(owner="andy@chalk.ai")

    ft: datetime = feature_time()


@features(tags=["1", "2"])
class TagFeatures:
    empty: str
    one: str = feature(tags="one")
    many: str = feature(tags=["a", "b"])

    ft: datetime = feature_time()


class ParseCommentsTestCase(unittest.TestCase):
    def test_parse_descriptions(self):
        self.assertEqual("bizarre", unwrap_feature(WowFS.bizarre).description)
        self.assertEqual(
            "This is a really neat description of something",
            unwrap_feature(WowFS.something).description,
        )
        self.assertEqual("explicit comment", unwrap_feature(WowFS.explicit).description)
        self.assertEqual(
            "This is also really neat and cool",
            unwrap_feature(WowFS.something_else).description,
        )
        self.assertEqual("now with feature", unwrap_feature(WowFS.assigned_comment).description)
        self.assertEqual("goofy", unwrap_feature(WowFS.goofy).description)
        self.assertEqual(
            """Multiline
Neat and, verily, cool

Hello""",
            unwrap_feature(WowFS.assigned_comment_multiline).description,
        )

    def test_class_owner(self):
        self.assertEqual("elliot@chalk.ai", OwnerFeatures.__chalk_owner__)
        self.assertEqual("elliot@chalk.ai", unwrap_feature(OwnerFeatures.plain).owner)
        self.assertEqual("elliot@chalk.ai", unwrap_feature(OwnerFeatures.cached).owner)
        self.assertEqual("andy@chalk.ai", unwrap_feature(OwnerFeatures.andy).owner)

    def test_class_tags(self):
        self.assertEqual(["1", "2"], TagFeatures.__chalk_tags__)
        self.assertEqual(["1", "2"], unwrap_feature(TagFeatures.empty).tags)
        self.assertEqual(["one", "1", "2"], unwrap_feature(TagFeatures.one).tags)
        self.assertEqual(["a", "b", "1", "2"], unwrap_feature(TagFeatures.many).tags)
        self.assertEqual(["a", "t"], unwrap_feature(WowFS.assigned).tags)
        self.assertEqual(["t"], unwrap_feature(WowFS.nope).tags)
        self.assertEqual(["3", "4", "t"], unwrap_feature(WowFS.assigned_comment).tags)


if __name__ == "__main__":
    unittest.main()



import itertools
import os
from pathlib import Path
from typing import Literal, Mapping, Optional

import yaml
from pydantic import BaseModel


class EnvironmentSettings(BaseModel):
    runtime: Optional[Literal["python310"]]
    requirements: Optional[str]
    dockerfile: Optional[str]


class ProjectSettings(BaseModel):
    project: str
    environments: Optional[Mapping[str, EnvironmentSettings]]
    local_path: str


def _load_project_config_at_path(filename: Path) -> Optional[ProjectSettings]:
    has_default_requirements = os.path.exists(filename.parent / "requirements.txt")
    try:
        with open(filename, "r") as f:
            parsed = yaml.safe_load(f)
            settings = ProjectSettings(
                **parsed,
                local_path=str(filename.absolute().resolve()),
            )
            if has_default_requirements:
                for i, cfg in settings.environments.items():
                    if cfg.requirements is None:
                        cfg.requirements = "requirements.txt"
            return settings
    except OSError:
        return None


def load_project_config() -> Optional[ProjectSettings]:
    base = Path(os.getcwd())

    for d in itertools.chain([base], base.parents):
        project = _load_project_config_at_path(d / "chalk.yaml") or _load_project_config_at_path(d / "chalk.yml")

        if project is not None:
            return project

    return None



import contextlib
import enum
from typing import Type

import pytest

from chalk.utils.enum import get_enum_value_type


class EmptyEnum(enum.Enum):
    pass


class ValidEnum(enum.Enum):
    A = 2
    B = 3


class IntFloatEnum(enum.Enum):
    A = 2
    B = 3.0


class FloatIntEnum(enum.Enum):
    A = 2.0
    B = 3


class InvalidEnum(enum.Enum):
    A = 3
    B = "hello"


@pytest.mark.parametrize(
    "enum_cls,expected_type,raises",
    [
        (EmptyEnum, None, False),
        (ValidEnum, int, False),
        (IntFloatEnum, float, False),
        (FloatIntEnum, float, False),
        (InvalidEnum, None, True),
    ],
)
def test_get_enum_value_type(enum_cls: Type[enum.Enum], expected_type: Type, raises: bool):
    with pytest.raises(TypeError) if raises else contextlib.nullcontext():
        assert get_enum_value_type(enum_cls) is expected_type



import json
import os
from typing import Optional


def has_integration(integration_name: str) -> bool:
    encoded = os.getenv("_CHALK_AVAILABLE_INTEGRATIONS")
    if encoded is not None:
        available = set(json.loads(encoded))
        return integration_name in available
    return False


def load_integration_variable(
    name: str,
    integration_name: Optional[str],
) -> Optional[str]:
    if integration_name is None:
        return os.getenv(name)
    elif has_integration(integration_name):
        return os.getenv(f"{integration_name}_{name}")
    return None



from typing import Any, Callable, Mapping, Optional, Protocol, Union

from pydantic import BaseModel
from sqlalchemy.engine.url import URL
from sqlalchemy.orm import Session

from chalk.features import Feature
from chalk.utils.duration import Duration


class IncrementalSettings(BaseModel):
    lookback_period: Optional[Duration]


class StringChalkQueryProtocol(Protocol):
    def execute(self):
        ...

    def one_or_none(self):
        ...

    def one(self):
        ...

    def all(self, incremental: Union[bool, IncrementalSettings] = False):
        ...


class ChalkQueryProtocol(Protocol):
    def first(self) -> "ChalkQueryProtocol":
        ...

    def one_or_none(self) -> "ChalkQueryProtocol":
        ...

    def one(self) -> Any:
        ...

    def all(self, incremental: Union[bool, IncrementalSettings] = False) -> Any:
        ...

    def filter_by(self, **kwargs) -> "ChalkQueryProtocol":
        ...

    def filter(self, *criterion) -> "ChalkQueryProtocol":
        ...

    def order_by(self, *clauses) -> "ChalkQueryProtocol":
        ...

    def group_by(self, *clauses) -> "ChalkQueryProtocol":
        ...

    def having(self, criterion) -> "ChalkQueryProtocol":
        ...

    def union(self, *q) -> "ChalkQueryProtocol":
        ...

    def union_all(self, *q) -> "ChalkQueryProtocol":
        ...

    def intersect(self, *q) -> "ChalkQueryProtocol":
        ...

    def intersect_all(self, *q) -> "ChalkQueryProtocol":
        ...

    def join(self, target, *props, **kwargs) -> "ChalkQueryProtocol":
        ...

    def outerjoin(self, target, *props, **kwargs) -> "ChalkQueryProtocol":
        ...

    def select_from(self, *from_obj) -> "ChalkQueryProtocol":
        ...

    def execute(self):
        ...


class DBSessionProtocol(Protocol):
    def update_query(self, f: Callable[[Session], Session]) -> None:
        ...

    def result(self) -> Any:
        ...

    def execute(self, q) -> Any:
        ...

    def close(self):
        ...


class DBSessionMakerProtocol(Protocol):
    def get_session(self, source: "BaseSQLSourceProtocol") -> DBSessionProtocol:
        ...


class BaseSQLSourceProtocol(Protocol):
    def query_string(
        self,
        query: str,
        fields: Mapping[str, Feature],
        args: Optional[Mapping[str, str]],
    ) -> StringChalkQueryProtocol:
        ...

    def query(self, *entities, **kwargs) -> ChalkQueryProtocol:
        ...

    def local_engine_url(self) -> Union[str, URL]:
        ...

    def set_session_maker(self, maker: DBSessionMakerProtocol) -> None:
        ...

    def engine_args(self) -> Mapping[str, Any]:
        return {}



from chalk.features.feature import Features
from chalk.features.feature_set import features


@features
class UserFeatures:
    name: str
    bday: str
    age: int


def name_age_bad() -> Features[UserFeatures.name, UserFeatures.age]:
    # We're getting Features[name, bday] and expected Features[name, age]
    return UserFeatures(name="4", bday="")


def name_age_good_flipped() -> Features[UserFeatures.age, UserFeatures.name]:
    # Flips name and age to match
    return UserFeatures(name="4", age=4)


def name_bad_type() -> Features[UserFeatures.name]:
    # Knows that name should be a string, not an int
    UserFeatures(what=4)
    return UserFeatures(name=4)



import unittest
from datetime import datetime

from chalk.features import features, has_one
from chalk.features.feature_set import feature_time


@features
class NoFunFeatures:
    id: str


@features
class FunFeatures:
    id: str
    nope: str
    single_parent: "NoFunFeatures" = has_one(lambda: FunFeatures.nope == NoFunFeatures.id)
    ts = feature_time()


class BackPopulationOfFeaturesTestCase(unittest.TestCase):
    def test_features(self):
        now = datetime.now()
        self.assertDictEqual(
            {"fun_features.nope": "hello", "fun_features.ts": now},
            dict(FunFeatures(nope="hello", ts=now)),
        )
        self.assertDictEqual(
            {"fun_features.nope": "hello"},
            dict(FunFeatures(nope="hello")),
        )
        self.assertDictEqual(
            {"fun_features.nope": "hello"},
            dict(FunFeatures(nope="hello", single_parent=NoFunFeatures(id="a"))),
        )
        self.assertDictEqual({}, dict(FunFeatures()))


if __name__ == "__main__":
    unittest.main()



from chalk.features.feature import Features
from chalk.features.feature_set import features


@features
class UserProfile:
    user_id: str
    profile_id: str
    address: int


@features
class Account:
    account_id: str
    balance: int


@features
class Transaction:
    user_id: str
    amount: int
    source_account_id: int
    dest_account_id: int


@features
class UserFeatures:
    uid: str
    name: str
    bday: str
    age: int


def name_age_good(
    age: UserFeatures.age,
) -> Features[UserFeatures.name, UserFeatures.age]:
    # Correctly knows that age is an int
    age_int_cast: int = age
    print(age_int_cast)
    # Correctly casts UserFeatures.__init__ to Features[name, age]
    return UserFeatures(name="4", age=4)


def name_age_good_flipped(
    age: UserFeatures.age,
) -> Features[UserFeatures.age, UserFeatures.name]:
    # Flips name and age to match
    return UserFeatures(name="4", age=4)


def name_age_bad() -> Features[UserFeatures.name, UserFeatures.age]:
    # We're getting Features[name, bday] and expected Features[name, age]
    return UserFeatures(name="4", bday="")


def name_bad_type() -> Features[UserFeatures.name]:
    # Knows that name should be a string, not an int
    return UserFeatures(name=4)



from typing import List, Union

# Tags allow you to scope requests within an
# environment. Both tags and environment need
# to match for a resolver to be a candidate to
# execute.
Tags = Union[List[str], str]

# Environments are used to trigger behavior
# in different deployments such as staging,
# production, and local development.
# For example, you may wish to interact with
# a vendor via an API call in the production
# environment, and opt to return a constant
# value in a staging environment.
#
# Environment can take one of three types:
#   - None (default) - candidate to run in every environment
#   - str - run only in this environment
#   - list[str] - run in any of the specified environment and no others
Environments = Union[List[str], str]



import dataclasses
import enum
from datetime import date, datetime, timezone
from typing import Any, Dict, Optional

import cattrs
import numpy as np
import pandas
import pendulum
import pydantic
import pytz
from dateutil import parser

from chalk.features import Feature, FeatureSetBase
from chalk.features.feature import Features
from chalk.utils.enum import get_enum_value_type

try:
    import attrs
except ImportError:
    # Imports not available. Attrs is not required.
    attrs = None


class FeatureCodec:
    def __init__(
        self,
        fqn_to_feature: Optional[Dict[str, Feature]] = None,
    ):
        self._fqn_to_feature = (
            fqn_to_feature
            if fqn_to_feature is not None
            else {
                feature.fqn: feature
                for fsb in FeatureSetBase.registry.values()
                for feature in fsb.features
                if isinstance(feature, Feature)
            }
        )
        self.converter = cattrs.Converter()
        self.converter.register_structure_hook(datetime, lambda v, _: parser.isoparse(v))
        self.converter.register_unstructure_hook(
            datetime, lambda v: (v if v.tzinfo else pytz.utc.localize(v)).isoformat()
        )

    def _default_encode(self, value: Any, for_pandas: bool = False):
        if isinstance(value, list):
            return [self._default_encode(x) for x in value]
        if isinstance(value, (str, int, float)):
            return value
        if isinstance(value, enum.Enum):
            return self._default_encode(value.value)
        if isinstance(value, datetime):
            tz = value.tzinfo or (datetime.now(timezone.utc).astimezone().tzinfo)
            return pendulum.instance(value, tz).isoformat()
        if isinstance(value, date):
            return value.isoformat()
        if isinstance(value, np.integer):
            return int(value)
        if isinstance(value, np.floating):
            return float(value)
        if isinstance(value, pandas.Timestamp):
            return pendulum.instance(value.to_pydatetime()).isoformat()
        if isinstance(value, pydantic.BaseModel):
            return value.dict()
        if (attrs is not None and attrs.has(type(value))) or dataclasses.is_dataclass(value):
            return self.converter.unstructure(value)
        raise TypeError(f"Unable to encode value of type {type(value).__name__}")

    def encode(
        self,
        feature: Feature,
        value: Any,
    ):
        if value is None:
            return None

        if feature.encoder is not None:
            return feature.encoder(value)

        return self._default_encode(value)

    def encode_fqn(self, fqn: str, value: Any):
        if fqn in self._fqn_to_feature:
            return self.encode(self._fqn_to_feature[fqn], value)

        return self._default_encode(value)

    def _default_decode_value(
        self,
        feature: Feature,
        value: Any,
    ):
        assert feature.typ is not None
        if issubclass(feature.typ.underlying, enum.Enum):
            value = feature.typ.underlying(value)
        if issubclass(feature.typ.underlying, datetime):
            value = parser.isoparse(value)
        elif issubclass(feature.typ.underlying, date):
            # note: datetime is a subclass of date, so we must be careful to decode accordingly
            value = date.fromisoformat(value)
        if issubclass(feature.typ.underlying, pydantic.BaseModel):
            return feature.typ.underlying(**value)
        elif (attrs is not None and attrs.has(feature.typ.underlying)) or dataclasses.is_dataclass(
            feature.typ.underlying
        ):
            return self.converter.structure(value, feature.typ.underlying)
        if not isinstance(value, feature.typ.underlying):
            raise TypeError(f"Unable to decode value {value} to type {feature.typ.underlying.__name__}")
        return value

    def _default_decode(
        self,
        feature: Feature,
        value: Any,
    ):
        assert feature.typ is not None
        if value is None:
            if feature.typ.is_nullable:
                return None
            else:
                raise ValueError(f"Value is none but feature {feature} is not nullable")
        if feature.typ.is_list:
            if not isinstance(value, list):
                raise TypeError(f"Feature {feature} is a list but value {value} is not")
            return [self._default_decode_value(feature, x) for x in value]
        else:
            return self._default_decode_value(feature, value)

    def decode(
        self,
        feature: Feature,
        value: Any,
    ):
        if value is None:
            return None

        if feature.decoder is not None:
            return feature.decoder(value)

        return self._default_decode(feature, value)

    def decode_fqn(
        self,
        fqn: str,
        value: Any,
    ):
        if fqn in self._fqn_to_feature:
            return self.decode(self._fqn_to_feature[fqn], value)
        return value

    def get_dtype(self, fqn: str) -> str:
        feature = self._fqn_to_feature[fqn]
        typ = feature.typ
        assert typ is not None, "typ should be specified"
        underlying = typ.underlying
        if issubclass(underlying, enum.Enum):
            # For enums, require all members to have the same type
            underlying = get_enum_value_type(underlying)
            assert underlying is not None, "enums should have at least one member"
        # See https://pandas.pydata.org/docs/user_guide/basics.html#basics-dtypes
        if issubclass(underlying, str):
            return "object"
        if issubclass(underlying, bool):
            return "boolean"
        if issubclass(underlying, int):
            return "Int64"
        if issubclass(underlying, float):
            return "Float64"
        if issubclass(underlying, datetime):
            return "object"
        if issubclass(underlying, date):
            return "object"
        if issubclass(underlying, Features):
            return "object"
        raise TypeError(f"Unsupported type: {underlying.__name__}")



import dataclasses
import inspect
from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Protocol, Type, Union, get_type_hints

from chalk.features import DataFrame
from chalk.features.feature import Feature, Features, FeatureWrapper, Filter
from chalk.features.tag import Environments, Tags
from chalk.sql.base.protocols import BaseSQLSourceProtocol
from chalk.utils.duration import Duration, ScheduleOptions

MachineType = str


class FilterFunction(Protocol):
    def __call__(self, *args: Feature) -> bool:
        ...


class SampleFunction(Protocol):
    def __call__(self) -> Features:
        ...


@dataclass
class Cron:
    schedule: Optional[ScheduleOptions] = None
    filter: Optional[FilterFunction] = None
    sample: Optional[SampleFunction] = None


class Resolver(Protocol):
    function_definition: str
    fqn: str
    filename: str
    inputs: List[Feature]
    output: Type[Features]
    fn: Callable
    environment: Optional[Environments]
    tags: Optional[Tags]
    max_staleness: Optional[Duration]
    machine_type: Optional[MachineType]

    registry: "List[Resolver]" = []

    def __hash__(self):
        return hash(self.fqn)


def _process_call(result):
    from chalk.sql.integrations.chalk_query import ChalkQuery, StringChalkQuery

    if isinstance(result, (ChalkQuery, StringChalkQuery)):
        return result.execute()

    return result


class SinkResolver:
    registry: "List[SinkResolver]" = []

    def __eq__(self, other):
        return isinstance(other, SinkResolver) and self.fqn == other.fqn

    def __hash__(self):
        return hash(self.fqn)

    def __call__(self, *args, **kwargs):
        return _process_call(self.fn(*args, **kwargs))

    def __init__(
        self,
        function_definition: str,
        fqn: str,
        filename: str,
        doc: Optional[str],
        inputs: List[Feature],
        output: Features,
        fn: Callable,
        environment: Optional[Environments],
        tags: Optional[Tags],
        machine_type: Optional[MachineType],
        buffer_size: Optional[int],
        debounce: Optional[Duration],
        max_delay: Optional[Duration],
        upsert: Optional[bool],
        integration: Optional[BaseSQLSourceProtocol] = None,
    ):
        self.function_definition = function_definition
        self.fqn = fqn
        self.filename = filename
        self.inputs = inputs
        self.output = output
        self.fn = fn
        self.environment = environment
        self.tags = tags
        self.doc = doc
        self.machine_type = machine_type
        self.buffer_size = buffer_size
        self.debounce = debounce
        self.max_delay = max_delay
        self.upsert = upsert
        self.integration = integration

    def __repr__(self):
        return f"SinkResolver(name={self.fqn})"


class OnlineResolver(Resolver):
    cron: Union[ScheduleOptions, Cron]

    def __eq__(self, other):
        return isinstance(other, OnlineResolver) and self.fqn == other.fqn

    def __hash__(self):
        return hash(self.fqn)

    def __call__(self, *args, **kwargs):
        return _process_call(self.fn(*args, **kwargs))

    def __init__(
        self,
        function_definition: str,
        fqn: str,
        filename: str,
        doc: Optional[str],
        inputs: List[Feature],
        output: Type[Features],
        fn: Callable,
        environment: Optional[Environments],
        tags: Optional[Tags],
        max_staleness: Optional[Duration],
        cron: Optional[Union[ScheduleOptions, Cron]],
        machine_type: Optional[MachineType],
        when: Optional[Filter],
    ):
        self.function_definition = function_definition
        self.fqn = fqn
        self.filename = filename
        self.inputs = inputs
        self.output = output
        self.fn = fn
        self.environment = environment
        self.tags = tags
        self.max_staleness = max_staleness
        self.cron = cron
        self.doc = doc
        self.machine_type = machine_type
        self.when = when

    def __repr__(self):
        return f"OnlineResolver(name={self.fqn})"


class OfflineResolver(Resolver):
    def __eq__(self, other):
        return isinstance(other, OfflineResolver) and self.fqn == other.fqn

    def __hash__(self):
        return hash(self.fqn)

    def __call__(self, *args, **kwargs):
        return _process_call(self.fn(*args, **kwargs))

    def __init__(
        self,
        function_definition: str,
        fqn: str,
        filename: str,
        doc: Optional[str],
        inputs: List[Feature],
        output: Type[Features],
        fn: Callable,
        environment: Optional[Environments],
        tags: Optional[Tags],
        max_staleness: Optional[Duration],
        cron: Union[ScheduleOptions, Cron],
        machine_type: Optional[MachineType],
    ):
        self.function_definition = function_definition
        self.fqn = fqn
        self.filename = filename
        self.doc = doc
        self.inputs = inputs
        self.output = output
        self.fn = fn
        self.environment = environment
        self.tags = tags
        self.max_staleness = max_staleness
        self.cron = cron
        self.machine_type = machine_type

    def __repr__(self):
        return f"OfflineResolver(name={self.fqn})"


@dataclasses.dataclass
class ResolverParseResult:
    fqn: str
    inputs: List[Feature]
    output: Optional[Type[Features]]
    function_definition: str
    function: Callable
    doc: Optional[str]


def parse_function(fn: Callable, ignore_return: bool = False) -> ResolverParseResult:
    fqn = f"{fn.__module__}.{fn.__name__}"
    sig = inspect.signature(fn)

    function_definition = inspect.getsource(fn)
    return_annotation = get_type_hints(fn)["return"]

    ret_val = None

    if isinstance(return_annotation, FeatureWrapper):
        return_annotation = return_annotation._chalk_feature

    if isinstance(return_annotation, Feature):
        assert return_annotation.typ is not None

        if return_annotation.is_has_many:
            assert issubclass(return_annotation.typ.parsed_annotation, DataFrame)
            ret_val = Features[return_annotation.typ.parsed_annotation.columns]
        elif return_annotation.is_has_one:
            assert issubclass(return_annotation.typ.underlying, Features)
            ret_val = Features[return_annotation.typ.underlying.features]
        else:
            # function annotated like def get_account_id(user_id: User.id) -> User.account_id
            ret_val = Features[return_annotation]

    if ret_val is None:
        if not isinstance(return_annotation, type):
            raise TypeError(f"return_annotation {return_annotation} of type {type(return_annotation)} is not a type")
        if issubclass(return_annotation, Features):
            # function annotated like def get_account_id(user_id: User.id) -> Features[User.account_id]
            # or def get_account_id(user_id: User.id) -> User:
            ret_val = return_annotation
        elif issubclass(return_annotation, DataFrame):
            # function annotated like def get_transactions(account_id: Account.id) -> DataFrame[Transaction]
            ret_val = Features[return_annotation]

    if ret_val is None and not ignore_return:
        raise ValueError(f"Resolver {fqn} did not have a valid return type")

    inputs = [p.annotation for p in sig.parameters.values()]

    # Unwrap anything that is wrapped with FeatureWrapper
    inputs = [p._chalk_feature if isinstance(p, FeatureWrapper) else p for p in inputs]

    return ResolverParseResult(
        fqn=fqn,
        inputs=inputs,
        output=ret_val,
        function_definition=function_definition,
        function=fn,
        doc=fn.__doc__,
    )


def online(
    fn: Optional[Callable] = None,
    /,
    environment: Optional[Environments] = None,
    tags: Optional[Tags] = None,
    cron: Optional[Union[ScheduleOptions, Cron]] = None,
    machine_type: Optional[MachineType] = None,
    when: Optional[Any] = None,
):
    caller_filename = inspect.stack()[1].filename

    def decorator(args, cf=caller_filename):
        parsed = parse_function(args)
        if parsed.fqn in {s.fqn for s in Resolver.registry}:
            raise ValueError(f"Duplicate resolver {parsed.fqn}")
        Resolver.registry.append(
            OnlineResolver(
                filename=cf,
                function_definition=parsed.function_definition,
                fqn=parsed.fqn,
                doc=parsed.doc,
                inputs=parsed.inputs,
                output=parsed.output,
                fn=parsed.function,
                environment=environment,
                tags=tags,
                max_staleness=None,
                cron=cron,
                machine_type=machine_type,
                when=when,
            )
        )
        return args

    return decorator(fn) if fn else decorator


def offline(
    fn: Optional[Callable] = None,
    environment: Optional[Environments] = None,
    tags: Optional[Tags] = None,
    cron: Union[ScheduleOptions, Cron] = None,
    machine_type: Optional[MachineType] = None,
):
    caller_filename = inspect.stack()[1].filename

    def decorator(args, cf=caller_filename):
        parsed = parse_function(args)
        if parsed.fqn in {s.fqn for s in Resolver.registry}:
            raise ValueError(f"Duplicate resolver {parsed.fqn}")
        Resolver.registry.append(
            OfflineResolver(
                filename=cf,
                function_definition=parsed.function_definition,
                fqn=parsed.fqn,
                doc=parsed.doc,
                inputs=parsed.inputs,
                output=parsed.output,
                fn=parsed.function,
                environment=environment,
                tags=tags,
                max_staleness=None,
                cron=cron,
                machine_type=machine_type,
            )
        )
        return args

    return decorator(fn) if fn else decorator


def sink(
    fn: Optional[Callable] = None,
    environment: Optional[Environments] = None,
    tags: Optional[Tags] = None,
    machine_type: Optional[MachineType] = None,
    buffer_size: Optional[int] = None,
    debounce: Optional[Duration] = None,
    max_delay: Optional[Duration] = None,
    upsert: Optional[bool] = None,
    integration: Optional[BaseSQLSourceProtocol] = None,
):
    caller_filename = inspect.stack()[1].filename

    def decorator(args, cf=caller_filename):
        parsed = parse_function(args, ignore_return=True)
        SinkResolver.registry.append(
            SinkResolver(
                filename=cf,
                function_definition=parsed.function_definition,
                fqn=parsed.fqn,
                doc=parsed.doc,
                inputs=parsed.inputs,
                output=parsed.output,
                fn=parsed.function,
                environment=environment,
                tags=tags,
                machine_type=machine_type,
                buffer_size=buffer_size,
                debounce=debounce,
                max_delay=max_delay,
                upsert=upsert,
                integration=integration,
            )
        )
        return args

    return decorator(fn) if fn else decorator



from __future__ import annotations

from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Mapping, Optional, Tuple, Type, Union, cast

import pandas as pd

from chalk.utils.collections import ensure_tuple, get_unique_item

if TYPE_CHECKING:
    from chalk.features.feature import Feature, Features, Filter


class DataFrameMeta(type):
    def __getitem__(cls, item) -> Type[DataFrame]:
        # leaving untyped as we type the individual features as their object type
        # but item should really be filter (expressions), Feature classes, or Feature instances
        from chalk.df.ChalkASTParser import ChalkASTParser
        from chalk.features.feature import Feature, Features, FeatureWrapper, Filter

        cls = cast(Type[DataFrame], cls)

        item = ensure_tuple(item)

        # Disallow string annotations like DataFrame["User"].
        # Instead, the entire thing should be in quotes -- like "DataFrame[User]"
        for x in item:
            if isinstance(x, str):
                raise TypeError(
                    f'Annotation {cls.__name__}["{x}", ...] is unsupported. Instead, use a string for the entire annotation -- for example: '
                    f'"{cls.__name__}[{x}, ...]"'
                )

        # If doing multiple subscript, then keep the filters, but do not keep the individual columns
        # TODO: Validate that any new columns are a subset of the existing columns
        item = [*item, *cls.filters]

        item = ChalkASTParser.parse_dataframe_getitem(item)

        new_filters: List[Filter] = []
        new_references_feature_set: Optional[Type[Features]] = None
        new_columns: List[Feature] = []

        for a in item:
            if isinstance(a, Filter):
                new_filters.append(a)
            elif isinstance(a, type) and issubclass(a, Features):
                if new_references_feature_set is not None:
                    raise ValueError(
                        f"Multiple referenced feature sets -- {new_references_feature_set} and {a} -- are not supported."
                    )
                new_references_feature_set = a
            elif isinstance(a, Feature):
                new_columns.append(a)
            elif isinstance(a, FeatureWrapper):
                new_columns.append(a._chalk_feature)
            else:
                raise TypeError(f"Invalid type for DataFrame[{a}]: {type(a)}")

        if len(new_columns) == 0 and new_references_feature_set is None:
            # This is possible if you have something like
            # Users.transactions[after('60d')]
            # In this case, keep all existing columns
            # But if you did
            # Users.transactions[Transaction.id, after('60d')]
            # Then keep only the id column
            new_columns = list(cls.columns)

        class SubclassedDataFrame(cls):
            filters = tuple(new_filters)
            __columns__ = tuple(new_columns)
            references_feature_set = new_references_feature_set

        return SubclassedDataFrame

    def __call__(cls, *args: None, **kwargs: None):
        # Allowing args and kwargs to display our custom error message below
        raise RuntimeError("DataFrame(...) should never be instantiated. Use DataFrame[...] notation instead.")

    def __repr__(cls):
        cls = cast(Type[DataFrame], cls)
        elements = [str(x) for x in (*cls.filters, *cls.columns)]
        if cls.references_feature_set is not None:
            elements.append(cls.references_feature_set.__name__)
        return f"DataFrame[{', '.join(elements)}]"

    def __instancecheck__(self, other: object) -> bool:
        raise RuntimeError(
            "DataFrame cannot be used with an isinstance(x, DataFrame) check. Instead, use issubclass(x, DataFame)."
        )

    @property
    def columns(cls) -> Tuple[Feature, ...]:
        # Computing the columns lazily as we need to implicitly parse the type annotation
        # to determine if a field is a has-many, and we don't want to do that on the
        # __getitem__ which could happen before forward references can be resolved
        # So, using a property on the metaclass, which acts like an attribute on the class, to
        # provide the dataframe columns
        from chalk.features.feature import Feature

        cls = cast(Type[DataFrame], cls)
        columns: List[Feature] = []
        for x in cls.__columns__:
            if not x.is_has_many:
                columns.append(x)
        if cls.references_feature_set is not None:
            for x in cls.references_feature_set.features:
                assert isinstance(x, Feature)
                if not x.is_has_many and not x.is_has_one:
                    columns.append(x)

        return tuple(columns)

    @property
    def namespace(cls) -> str:
        cls = cast(Type[DataFrame], cls)
        namespaces = [x.path[0].parent.namespace if len(x.path) > 0 else x.namespace for x in cls.columns]
        return get_unique_item(namespaces, f"dataframe {cls.__name__} column namespaces")


class DataFrame(metaclass=DataFrameMeta):
    filters: Tuple[Filter, ...] = ()
    columns: Tuple[Feature, ...]  # set via a decorator on the metaclass
    __columns__: Tuple[Feature, ...] = ()
    references_feature_set: Optional[Type[Features]] = None

    @classmethod
    def from_dict(cls, d: Dict[Any, List[Any]]):
        from chalk.df.ChalkDataFrameImpl import ChalkDataFrameImpl
        from chalk.parsed.feature import FeatureType

        return ChalkDataFrameImpl.from_pandas(pd.DataFrame({FeatureType.of(k): v for k, v in d.items()}))

    @classmethod
    def from_list(cls, *args: Union[List[Any], Any]):
        from chalk.features.feature import Features

        def _yield_args(a):
            if isinstance(a, Features):
                yield a

            elif isinstance(a, Iterable):
                for aa in a:
                    yield from _yield_args(aa)

            else:
                raise ValueError("Items in list must be wrapped in a feature set")

        from_dict = defaultdict(list)
        keys_frozen = False
        for arg in _yield_args(args):
            arg_dict = dict(arg)
            if keys_frozen and arg_dict.keys() != from_dict.keys():
                raise ValueError(
                    f"DataFrame.from_list received objects with different sets of features:"
                    f" {set(arg_dict.keys())} vs {set(from_dict.keys())}"
                )

            for k, v in arg_dict.items():
                from_dict[k].append(v)

            keys_frozen = True

        return cls.from_dict(from_dict)

    @classmethod
    def from_pandas(
        cls,
        df: pd.DataFrame,
    ):
        from chalk.df.ChalkDataFrameImpl import ChalkDataFrameImpl

        return ChalkDataFrameImpl.from_pandas(df)

    @classmethod
    def read_parquet(
        cls,
        path: Union[str, Path],
        columns: Union[Mapping[str, Any], Mapping[int, Any]],
    ):
        from chalk.df.ChalkDataFrameImpl import ChalkDataFrameImpl
        from chalk.parsed.feature import FeatureType

        return ChalkDataFrameImpl.from_pandas(
            pd.read_parquet(
                path=path,
                columns=list(columns.keys()),
            ).rename(columns={k: FeatureType.of(v) for k, v in columns.items()})
        )

    @classmethod
    def read_csv(
        cls,
        path: Union[str, Path],
        columns: Union[Mapping[str, Any], Mapping[int, Any]],
    ):
        from chalk.df.ChalkDataFrameImpl import ChalkDataFrameImpl
        from chalk.parsed.feature import FeatureType

        return ChalkDataFrameImpl.from_pandas(
            pd.read_csv(filepath_or_buffer=path).rename(columns={k: FeatureType.of(v) for k, v in columns.items()})
        )



import datetime
import enum
import functools
import logging
from typing import Any, Union

import pandas as pd

from chalk.features import Feature
from chalk.features.feature import FeatureWrapper, Filter, TimeDelta, unwrap_feature
from chalk.parsed.feature import FeatureTimeFeatureType, FeatureType

_logger = logging.getLogger(__name__)


def _feature_type_or_value(e):
    if isinstance(e, (Feature, FeatureWrapper)):
        return FeatureType.of(e)
    return e


class ChalkDataFrameImpl:
    def __init__(self, pandas_dataframe: pd.DataFrame):
        self.underlying = pandas_dataframe
        self.timestamp_feature = None
        for col in pandas_dataframe.columns:
            if isinstance(col, FeatureTimeFeatureType):
                if self.timestamp_feature is not None:
                    raise ValueError("DataFrame cannot have multiple timestamp features")
                self.timestamp_feature = col

    def _expect_scalar(self, v, name: str):
        if isinstance(v, pd.Series) and len(v) == 1:
            return v[0]
        if isinstance(v, pd.Series):
            raise ValueError(
                f"Cannot compute {name}. DataFrame contains {len(v)} features, and expected only one. "
                f"Filter your DataFrame down to the feature for which you want to compute the {name} "
                f"before calling .{name}()."
            )
        return v

    def mean(self):
        return self._expect_scalar(self.underlying.mean(), "mean")

    def median(self):
        return self._expect_scalar(self.underlying.median(), "median")

    def sum(self):
        return self._expect_scalar(self.underlying.sum(), "sum")

    def max(self):
        return self._expect_scalar(self.underlying.max(), "max")

    def min(self):
        return self._expect_scalar(self.underlying.min(), "min")

    def count(self) -> int:
        if isinstance(self.underlying, pd.DataFrame):
            return len(self.underlying)
        return self.underlying.count()

    def keys(self):
        return self.underlying.keys()

    def iterrows(self):
        return self.underlying.iterrows()

    @property
    def iloc(self):
        return self.underlying.iloc()

    def __repr__(self):
        return repr(self.underlying)

    def _maybe_replace_timestamp_feature(self, f: Union[Feature, Any]):
        """Replace the ``CHALK_TS`` pseudo-feature with the actual timestamp column."""
        if not isinstance(f, FeatureType) or f.fqn != "__chalk__.CHALK_TS":
            return f
        if self.timestamp_feature is None:
            raise ValueError("DataFrame has no timestamp")
        return self.timestamp_feature

    def _maybe_convert_timedelta_to_timestamp(
        self, f: Union[TimeDelta, datetime.timedelta, Any], now: datetime.datetime
    ):
        """Convert timedeltas relative to ``now`` into absolute datetimes."""
        if isinstance(f, TimeDelta):
            f = f.to_std()
        if isinstance(f, datetime.timedelta):
            return now + f
        return f

    def _parse_feature_or_value(self, f: Union[Feature, Any], now: datetime.datetime):
        """Parse a feature or value into the correct type that can be used for filtering."""
        f = _feature_type_or_value(f)
        f = self._maybe_convert_timedelta_to_timestamp(f, now)
        f = self._maybe_replace_timestamp_feature(f)
        if isinstance(f, enum.Enum):
            f = f.value
        return f

    def _get_filter(self, f: Filter, underlying: pd.DataFrame, now: datetime.datetime):
        """Evaluate a filter.

        Args:
            f (Filter): The filter.
            underlying (pd.DataFrame): The data frame.
            now (datetime.datetime): The datetime to use for the current timestamp. Used to resolve relative
                timestamps in filters to absolute datetimes.
        """
        # Passing `now` in explicitly instead of using datetime.datetime.now() so that multiple filters
        # relying on relative timestamps (e.g. before, after) will have the same "now" time.
        if f.operation == "not":
            return ~self._get_filter(f.feature, underlying, now)

        lhs = self._parse_feature_or_value(f.feature, now)
        rhs = self._parse_feature_or_value(f.other, now)

        if rhs is None and isinstance(lhs, FeatureType):
            if f.operation == "==":
                return underlying[lhs].isnull()

            elif f.operation == "!=":
                return underlying[lhs].notnull()

        if isinstance(lhs, FeatureType):
            lhs = underlying[lhs]

        if isinstance(rhs, FeatureType):
            rhs = underlying[rhs]

        if f.operation == "==":
            return lhs == rhs
        elif f.operation == "!=":
            return lhs != rhs
        elif f.operation == ">=":
            return lhs >= rhs
        elif f.operation == ">":
            return lhs > rhs
        elif f.operation == "<":
            return lhs < rhs
        elif f.operation == "<=":
            return lhs <= rhs
        elif f.operation == "in":
            return lhs.isin(rhs)
        elif f.operation == "not in":
            return ~lhs.isin(rhs)
        elif f.operation == "and":
            return self._get_filter(lhs, underlying, now) & self._get_filter(rhs, underlying, now)
        elif f.operation == "or":
            return self._get_filter(lhs, underlying, now) | self._get_filter(rhs, underlying, now)
        else:
            raise ValueError(f'Unknown operation "{f.operation}"')

    def __getitem__(self, item):
        try:
            from chalk.df.ChalkASTParser import ChalkASTParser

            item = ChalkASTParser.parse_dataframe_getitem(item)

            projections = []
            filters = []
            for x in item:
                if isinstance(x, FeatureWrapper):
                    x = unwrap_feature(x)
                if isinstance(x, Feature):
                    x = FeatureType.of(x)

                if isinstance(x, FeatureType):
                    projections.append(x)

                elif isinstance(x, Filter):
                    filters.append(x)

                else:
                    raise ValueError(f'Indexed chalk.DataFrame with invalid type "{type(item)}": {item}')
            underlying = self.underlying
            now = datetime.datetime.now()
            if len(filters) > 0:
                pandas_filters = [self._get_filter(f, underlying, now) for f in filters]
                single_filter = functools.reduce(lambda a, b: a & b, pandas_filters)
                underlying = underlying[single_filter]

            # Do the projection
            underlying = underlying[projections] if len(projections) > 0 else underlying

            return ChalkDataFrameImpl(underlying)

        except KeyError as e:
            _logger.debug(f"Failed to get key: {item}. Had keys: {self.underlying.keys()}", e)
            raise e

    @classmethod
    def from_pandas(cls, param):
        renames = {c: FeatureType.of(c) for c in param.columns}
        underlying = param.rename(columns=renames)
        return cls(underlying)

    def to_pandas(self) -> Union[pd.DataFrame, pd.Series]:
        renames = {c: str(c) for c in self.underlying.columns}
        return self.underlying.rename(columns=renames)
